/*
 * Copyright (c) 2010, 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: genhash.c,v 1.67 2014/01/30 16:04:23 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/stat.h"
#include "sm/fcntl.h"
#include "sm/ctype.h"
#include "sm/io.h"
#include "sm/string.h"
#include "sm/mta.h"
#include "sm/map.h"
#include "sm/dkim.h"
#include "sm/dkimdef.h"
#include "sm/base64.h"
#include "sm/mail_parse.h"

#include "dkim.h"

#if MTA_USE_DKIM

/* states for header parsing */
#define DKIM_HDR_BEGIN	0
#define DKIM_HDR_NAME	1		/* header name */
#define DKIM_HDR_COLON	(-1)	/* colon to separate name and value */
#define DKIM_HDR_VAL0	(-2)	/* right after colon */
#define DKIM_HDR_VAL	(-3)	/* header value */

#define ISWSP(c)	(' ' == (c) || '\t' == (c))

/* states for From: header collection */
#define FRMHDR_ST_NONE	0
#define FRMHDR_ST_NEW	(-1)	/* found From: header */
#define FRMHDR_ST_FIRST	(-2)	/* found From: header, str allocated */
#define FRMHDR_ST_SKIP	(-3)	/* skipping over initial white space */
#define FRMHDR_ST_RDING	(-4)	/* reading From: header */
#define FRMHDR_ST_GOT	(-5)	/* got From: header */
#define FRMHDR_ST_NOMEM	(-9)	/* can't allocate memory for From header */

/*
**  DKIM_PARSE_FROM -- parse From: header
**
**	Parameters:
**		dkim_msg -- DKIM msg context
**		c -- current header char
**		hdr_state -- header state (DKIM_HDR_*)
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_parse_from(dkim_msg_P dkim_msg, uchar c, int hdr_state)
{
	static const char fromhdr[] = "\r\nfrom:";

	if (FRMHDR_ST_GOT == dkim_msg->dkm_fhdr_state)
		return SM_SUCCESS;

	if (NULL == dkim_msg->dkm_fromhdr && dkim_msg->dkm_fhdr_state >= 0) {
		if (TOLOWER(c) == fromhdr[dkim_msg->dkm_fhdr_state]) {
			if (++dkim_msg->dkm_fhdr_state >= (int)strlen(fromhdr))
				dkim_msg->dkm_fhdr_state = FRMHDR_ST_NEW;
		}
		else {
			dkim_msg->dkm_fhdr_state = 0;
			if (TOLOWER(c) == fromhdr[dkim_msg->dkm_fhdr_state])
				++dkim_msg->dkm_fhdr_state;
		}
	}

	if (FRMHDR_ST_NEW == dkim_msg->dkm_fhdr_state) {
		if (NULL == dkim_msg->dkm_fromhdr)
			dkim_msg->dkm_fromhdr = sm_str_new(NULL, 80, 4096);
		else
			sm_str_clr(dkim_msg->dkm_fromhdr);
		if (dkim_msg->dkm_fromhdr != NULL)
			dkim_msg->dkm_fhdr_state = FRMHDR_ST_FIRST;
		else
			dkim_msg->dkm_fhdr_state = FRMHDR_ST_NOMEM;
	}

	if (FRMHDR_ST_FIRST == dkim_msg->dkm_fhdr_state)
		dkim_msg->dkm_fhdr_state = FRMHDR_ST_SKIP;
	else if (FRMHDR_ST_SKIP == dkim_msg->dkm_fhdr_state && ISSPACE(c))
		;
	else if (FRMHDR_ST_SKIP == dkim_msg->dkm_fhdr_state && ISPRINT(c))
		dkim_msg->dkm_fhdr_state = FRMHDR_ST_RDING;
	if (FRMHDR_ST_RDING == dkim_msg->dkm_fhdr_state) {
		if (DKIM_HDR_NAME == hdr_state) {
			dkim_msg->dkm_fhdr_state = FRMHDR_ST_GOT;
			sm_str_term(dkim_msg->dkm_fromhdr);
		}
		else if (ISPRINT(c))	/* all chars? */
			SM_STR_PUT(dkim_msg->dkm_fromhdr, c);
	}
	return SM_SUCCESS;
}

/*
**  DKIM_GET_SOPTS -- get signing options from configuration
**
**	Parameters:
**		dkim_ctx -- DKIM context
**		dkim_msg -- DKIM msg context
**		fromuser -- local part of From: address
**		fromdomain -- domain part of From: address
**		pkdim_s -- dkim signing context to use (output)
**
**	Returns:
**		SM_SUCCESS
*/

static sm_ret_T
dkim_get_sopts(dkim_ctx_P dkim_ctx, dkim_msg_P dkim_msg, sm_str_P fromuser, sm_str_P fromdomain, dkim_s_P *pdkim_s)
{
	sm_ret_T ret;
	sm_str_T rhs;

	if (NULL == dkim_ctx->dkc_dkim_map)
		return SM_SUCCESS;

	sm_str_assign(rhs, NULL, NULL, sizeof(void *), sizeof(void *));
	ret = sm_map_lookup_addr(dkim_ctx->dkc_dkim_map, fromuser, NULL,
			fromdomain, NULL, '+',
			SMMAP_LFL_FULL|SMMAP_LFL_USER|SMMAP_LFL_DOMAIN|SMMAP_LFL_LOCAL|SMMAP_LFL_DOTSUBDOM|SMMAP_LFL_USE_PTR,
			&rhs);
	/* sm_io_fprintf(smioerr, "sev=DBG, func=dkim_get_sopts, user=%S, domain=%S, ret=%r\n", fromuser, fromdomain, ret); */
	if (!sm_is_success(ret))
		return SM_SUCCESS;

	*pdkim_s = (dkim_s_P)sm_str_data(&rhs);
#if 0
sm_io_fprintf(smioerr, "sev=DBG, func=dkim_get_sopts, key=%S, dkim_s=%p\n", fromaddr, *pdkim_s);
#endif
	return SM_SUCCESS;
}

/*
**  DKIM_SKIP_HDR -- skip a header?
**
**	Parameters:
**		dkim_cnf -- DKIM conf context
**		hdrname -- name of header
**
**	Returns:
**		true iff header should not be signed (be skipped)
**
**  sign H iff (((Sign is empty) || (H is in Sign)) && (H is not in Skip))
**	=
**  skip H iff !(((Sign is empty) || (H is in Sign)) && (H is not in Skip))
**  skip H iff (!((Sign is empty) || (H is in Sign)) || (H is in Skip))
**  skip H iff (((Sign is not empty) && (H is not in Sign)) || (H is in Skip))
*/

static bool
dkim_skip_hdr(dkim_cnf_P dkim_cnf, sm_str_P hdrname)
{
	uint u;
	char *h;

	/* is header name listed in skip? */
	for (u = 0;
		(h = dkim_cnf->dkimcnf_skip[u]) != NULL &&
		u < SM_ARRAY_SIZE(dkim_cnf->dkimcnf_skip);
		u++)
	{
		if (strncasecmp(h, (const char *)sm_str_data(hdrname), strlen(h)) == 0)
			return true;
	}

	if (NULL == dkim_cnf->dkimcnf_sign[0])
		return false;

	for (u = 0;
		(h = dkim_cnf->dkimcnf_sign[u]) != NULL &&
		u < SM_ARRAY_SIZE(dkim_cnf->dkimcnf_sign);
		u++)
	{
		if (strncasecmp(h, (const char *)sm_str_data(hdrname), strlen(h)) == 0)
			return false;
	}

	return true;
}

/*
**  DKIM_ADD_HDRS -- add header names to signature
**
**	Parameters:
**		dkim_cnf -- DKIM conf context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
dkim_add_hdrs(dkim_cnf_P dkim_cnf, dkim_msg_P dkim_msg)
{
	sm_ret_T ret;
	uint u;
	char *hdradd;

	ret = SM_SUCCESS;
	if (NULL == dkim_msg->dkm_hdrs)
		return ret;

	/* add all header names that are not yet part of the signature */
	for (u = 0;
		(hdradd = dkim_cnf->dkimcnf_add[u]) != NULL &&
		u < SM_ARRAY_SIZE(dkim_cnf->dkimcnf_add);
		u++)
	{
		bool addit;
		size_t len;
		dkh_P dkh;

		len = strlen(hdradd);
		addit = true;
		DKH_FOREACH_REV(dkh, &(dkim_msg->dkm_hhd)) {
			const char *hdr;

			hdr = (const char *)sm_str_getdata(dkh->dkh_hdr_n);
			if (NULL == hdr) goto enomem;
			if (0 == strncasecmp(hdradd, hdr, len)) {
				addit = false;
				break;
			}
		}
		if (addit) {
			ret = sm_str_scat(dkim_msg->dkm_hdrs, hdradd);
			if (sm_is_err(ret)) goto error;
			ret = sm_str_put(dkim_msg->dkm_hdrs, ':');
			if (sm_is_err(ret)) goto error;
		}
	}
	return ret;

  enomem:
	ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
 error:
	(void) sm_str_rm_trail(dkim_msg->dkm_hdrs, ":");
	return ret;
}

/*
**  DKIM_MSG_CHUNK -- compute msg hashes
**
**	Parameters:
**		dkim_msg -- DKIM msg context
**		msg -- msg
**		len -- len of msg
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
dkim_msg_chunk(dkim_msg_P dkim_msg, const uchar *msg, size_t len)
{
	sm_ret_T ret;
	size_t i, hashlen;
	uchar *hash;
	uchar c;
	dkim_cookie_P dkim_cookie;
	dkim_ctx_P dkim_ctx;
	dkim_s_P dkim_s;
	dkim_cnf_P dkim_cnf;
	sm_str_P fromuser, fromdomain;
	static SM_DECL_EOT;
	static SM_DECL_EOH;
	bool maybeeot, eol;
#define SM_CRLFCR	(SM_EOH_LEN - 1)

#define DK_XFER_HDR_V	do { \
		if (dkim_msg->dkm_dkh != NULL) {	\
			dkim_msg->dkm_hdrtlen += sm_str_getlen(dkim_msg->dkm_hdr_v);	\
			if (dkim_msg->dkm_hdrtlen >= dkim_cnf->dkimcnf_hdrtlen) {	\
				ret = sm_error_perm(SM_EM_DKIM, SM_E_OVFLW_NS);	\
				goto error;	\
			}	\
			dkim_msg->dkm_dkh->dkh_hdr_v = sm_str_dup(NULL, dkim_msg->dkm_hdr_v); \
			if (NULL == dkim_msg->dkm_dkh->dkh_hdr_v) goto enomem;	\
			sm_str_clr(dkim_msg->dkm_hdr_v);	\
			dkim_msg->dkm_dkh = NULL;	\
		}	\
	} while (0)

	dkim_cookie = (dkim_cookie_P)f_cookie(*dkim_msg->dkm_hfp);
	dkim_ctx = dkim_msg->dkm_ctx;
	SM_ASSERT(dkim_ctx != NULL);
	dkim_cnf = dkim_ctx->dkc_cnf;
	SM_ASSERT(dkim_cnf != NULL);
	fromuser = fromdomain = NULL;
	eol = false;

	for (i = 0; i < len; i++) {
		c = msg[i];

		/*
		 *  header
		 *  states:
		 *  eoh: keeps track of CRLFCRLF and hence can be used to
		 *       - detect the end of a line (and thus the start of a header)
		 *       - detect the end of the headers
		 *       - avoid to include the 2nd CRLF
		 *  hdr: keeps track whether we are between CRLF and colon,
		 *       as well as directly behind the colon, thus allowing us to
		 *       - recognize and collect header names
		 *       - lower case header names [relaxed]
		 *       - remove white space around colon [relaxed]
		 */

		if (dkim_msg->dkm_eoh_state < SM_EOH_LEN) {
			bool skip;

			skip = false;
			if (c == eoh[dkim_msg->dkm_eoh_state]) {
				++dkim_msg->dkm_eoh_state;
				if (SM_EOH_LEN == dkim_msg->dkm_eoh_state) {
					dkh_P dkh;

					if (DKIM_C_REL(dkim_cnf->dkimcnf_hcanon) &&
					    !dkim_msg->dkm_skiphdr)
					{
						ret = sm_str_put(dkim_msg->dkm_hdr_v, '\r');
						if (sm_is_err(ret)) goto error;
						ret = sm_str_put(dkim_msg->dkm_hdr_v, '\n');
						if (sm_is_err(ret)) goto error;
					}
					dkim_msg->dkm_wsp = 0;

					DK_XFER_HDR_V;

					if (DKIM_PARSEFROM(dkim_msg) &&
						FRMHDR_ST_RDING == dkim_msg->dkm_fhdr_state)
					{
						dkim_msg->dkm_fhdr_state = FRMHDR_ST_GOT;
						sm_str_term(dkim_msg->dkm_fromhdr);
					}

					DKH_FOREACH_REV(dkh, &(dkim_msg->dkm_hhd)) {
						sm_io_fprintf(dkim_msg->dkm_hfp, "%S%S", dkh->dkh_hdr_n, dkh->dkh_hdr_v);
#if DKIM_DEBUG
						sm_io_fprintf(smioerr, "%@S%@S\n", dkh->dkh_hdr_n, dkh->dkh_hdr_v);
#endif
					}

					/* open DKIM buffer for body */
					ret = dkim_open_write(&dkim_ctx, DKIM_SIGNM(dkim_cnf), &dkim_msg->dkm_bfp);
					if (sm_is_err(ret)) goto error;
					continue;
				}
				if (dkim_msg->dkm_eoh_state == SM_CRLFCR)
					continue;
			}
			else {
				dkim_msg->dkm_eoh_state = 0;
				if (c == eoh[dkim_msg->dkm_eoh_state])
					++dkim_msg->dkm_eoh_state;
			}

			/* header name? */
			if (DKIM_HDR_BEGIN == dkim_msg->dkm_hdr_state && ISWSP(c))
				dkim_msg->dkm_hdr_state = DKIM_HDR_VAL;
			else if (dkim_msg->dkm_hdr_state >= 0) {
				if (!DKIM_C_REL(dkim_cnf->dkimcnf_hcanon))
					DK_XFER_HDR_V;

				if (DKIM_C_REL(dkim_cnf->dkimcnf_hcanon)) {
					if (ISSPACE(c))	/* skip over white space in header name */
						continue;
					c = TOLOWER(c);
				}

				if (':' == c) {
					dkim_msg->dkm_hdr_state = DKIM_HDR_COLON;
					ret = sm_str_put(dkim_msg->dkm_hdr_n, c);
					if (sm_is_err(ret)) goto error;
					dkim_msg->dkm_skiphdr = dkim_skip_hdr(dkim_cnf, dkim_msg->dkm_hdr_n);
					if (!dkim_msg->dkm_skiphdr) {
						skip = true;	/* don't write ':' again */

						ret = dkh_new(&(dkim_msg->dkm_hhd), &(dkim_msg->dkm_dkh));
						if (sm_is_err(ret)) goto error;
						dkim_msg->dkm_hdrtlen += sm_str_getlen(dkim_msg->dkm_hdr_n);
						if (dkim_msg->dkm_hdrtlen >= dkim_cnf->dkimcnf_hdrtlen) {
							ret = sm_error_perm(SM_EM_DKIM, SM_E_OVFLW_NS);
							goto error;
						}
						dkim_msg->dkm_dkh->dkh_hdr_n = sm_str_dup(NULL, dkim_msg->dkm_hdr_n);
						if (NULL == dkim_msg->dkm_dkh->dkh_hdr_n) goto enomem;
					}
					/* done with header name, next is header value */
					sm_str_clr(dkim_msg->dkm_hdr_n);
				}
				else {
					dkim_msg->dkm_hdr_state = DKIM_HDR_NAME;
					ret = sm_str_put(dkim_msg->dkm_hdr_n, c);
					if (sm_is_err(ret)) goto error;

					/* don't write header to hash yet... */
					skip = true;
				}
			}
			else if (eol = (dkim_msg->dkm_eoh_state == SM_EOL_LEN))
				dkim_msg->dkm_hdr_state = DKIM_HDR_BEGIN;
			if (dkim_msg->dkm_skiphdr)
				continue;

			/* convert multiple WS to a single space, preserve CRLF */
			if (DKIM_C_REL(dkim_cnf->dkimcnf_hcanon)) {
				if (ISSPACE(c)) {
					++dkim_msg->dkm_wsp;
					continue;
				}
				if (dkim_msg->dkm_wsp > 0) {
					if (eol) {
						ret = sm_str_put(dkim_msg->dkm_hdr_v, '\r');
						if (sm_is_err(ret)) goto error;
						ret = sm_str_put(dkim_msg->dkm_hdr_v, '\n');
						if (sm_is_err(ret)) goto error;
						DK_XFER_HDR_V;
					}
					else if (dkim_msg->dkm_hdr_state != DKIM_HDR_COLON) {
						ret = sm_str_put(dkim_msg->dkm_hdr_v, ' ');
						if (sm_is_err(ret)) goto error;
					}
				}
				dkim_msg->dkm_wsp = 0;
			}
			if (DKIM_HDR_COLON == dkim_msg->dkm_hdr_state && c != ':')
				dkim_msg->dkm_hdr_state = DKIM_HDR_VAL0;

			/* note: here we might get a modified header: no WS */
			if (DKIM_PARSEFROM(dkim_msg))
				dkim_parse_from(dkim_msg, c, dkim_msg->dkm_hdr_state);

			if (skip)
				continue;
			if (dkim_msg->dkm_hdr_state <= DKIM_HDR_VAL0 || eol) {
				ret = sm_str_put(dkim_msg->dkm_hdr_v, c);
				if (sm_is_err(ret)) goto error;
			}
			continue;
		}

		/*
		 *  body
		 *  states:
		 *  eot: keeps track of CRLF.CRLF and hence can be used to
		 *       - detect the end of a line
		 *       - detect the end of the message
		 *  el: keeps track of CRLFCRLF and hence can be used to
		 *       - detect empty lines; they are counted in elines
		 *  wsp: counts white space (space, tab) [relaxed]
		 */

		if (maybeeot = (c == eot[dkim_msg->dkm_eot_state])) {
			if (++dkim_msg->dkm_eot_state >= SM_EOT_LEN)
				break;
		}
		else {
			dkim_msg->dkm_eot_state = 0;
			if (c == eot[dkim_msg->dkm_eot_state])
				++dkim_msg->dkm_eot_state;
		}

		/* empty line? */
		if (c == eoh[dkim_msg->dkm_el_state]) {
			++dkim_msg->dkm_el_state;
			if (SM_EOL_LEN == dkim_msg->dkm_el_state) {
				++dkim_msg->dkm_elines;
				dkim_msg->dkm_el_state = 0;
			}
			dkim_msg->dkm_wsp = 0;
			continue;
		}
		else {
			if (!maybeeot && dkim_msg->dkm_elines > 0) {
				while (dkim_msg->dkm_elines-- > 0) {
					ret = sm_io_putc(dkim_msg->dkm_bfp, '\r');
					if (sm_is_err(ret)) goto error;
					ret = sm_io_putc(dkim_msg->dkm_bfp, '\n');
					if (sm_is_err(ret)) goto error;
				}
				dkim_msg->dkm_elines = 0;
			}
			dkim_msg->dkm_el_state = 0;
			if (c == eoh[dkim_msg->dkm_el_state])
				++dkim_msg->dkm_el_state;
		}

		/* skip leading '.' (can we move this up?) */
		if (dkim_msg->dkm_eot_state == SM_CRLFCR && c == '.') {
			dkim_msg->dkm_wsp = 0;
			continue;
		}

		if (DKIM_C_REL(dkim_cnf->dkimcnf_bcanon)) {
			if (ISWSP(c)) {
				++dkim_msg->dkm_wsp;
				continue;
			}
			else if (0 == dkim_msg->dkm_el_state && dkim_msg->dkm_wsp > 0) {
				ret = sm_io_putc(dkim_msg->dkm_bfp, ' ');
				if (sm_is_err(ret)) goto error;
			}
			dkim_msg->dkm_wsp = 0;
		}
		ret = sm_io_putc(dkim_msg->dkm_bfp, c);
		if (sm_is_err(ret)) goto error;
	}

	/* return if we reached end of chunk, but not eot */
	if (dkim_msg->dkm_eot_state < SM_EOT_LEN)
		return SM_NEEDMORE;

	if (dkim_msg->dkm_elines > 0) {
		ret = sm_io_putc(dkim_msg->dkm_bfp, '\r');
		if (sm_is_err(ret)) goto error;
		ret = sm_io_putc(dkim_msg->dkm_bfp, '\n');
		if (sm_is_err(ret)) goto error;
		dkim_msg->dkm_elines = 0;
	}

	/* store body hash in msg context */
	dkim_gethash(dkim_msg->dkm_bfp, &hash, &hashlen);
	sm_memzero(dkim_msg->dkm_b64bhash, sizeof(dkim_msg->dkm_b64bhash));
	sm_base64_encode(hash, hashlen, dkim_msg->dkm_b64bhash,
		sizeof(dkim_msg->dkm_b64bhash));

	ret = dkim_close(dkim_msg->dkm_bfp, 0);
	if (sm_is_err(ret)) goto error;
	dkim_msg->dkm_bfp = NULL;

	dkim_s = dkim_msg->dkm_s;
	if (DKIM_PARSEFROM(dkim_msg)) {
		uint addrlen;
		uchar *hdr, *user, *domain;
		dkim_s_P dkim_s_m;

		/* select right key etc... */
		if (NULL == dkim_msg->dkm_fromhdr ||
			sm_str_getlen(dkim_msg->dkm_fromhdr) < 3)
		{
			ret = sm_error_temp(SM_EM_DKIM, EINVAL);
			goto error;
		}
		hdr = sm_str_getdata(dkim_msg->dkm_fromhdr);
		if (NULL == hdr) goto enomem;
		ret = rfc2822_mailbox_split(hdr, &user, &domain);
		if (sm_is_err(ret)) goto error;
		if (NULL != user || NULL != domain) {
			if (NULL != user) {
				addrlen = (user != NULL) ? (uint)strlen((const char *)user) : 1u;
				fromuser = sm_str_new(NULL, addrlen + 2, addrlen + 8);
				if (NULL == fromuser) goto enomem;
				ret = sm_str_scopy(fromuser, (const char *)user);
				if (sm_is_err(ret)) goto error;
			}
			else
				fromuser = NULL;
			if (NULL != domain) {
				addrlen = (uint)strlen((const char *)domain);
				fromdomain = sm_str_new(NULL, addrlen + 2, addrlen + 8);
				if (NULL == fromdomain) goto enomem;
				ret = sm_str_scopy(fromdomain, (const char *)domain);
				if (sm_is_err(ret)) goto error;
			}
			else
				fromdomain = NULL;
			dkim_s_m = NULL;
			ret = dkim_get_sopts(dkim_ctx, dkim_msg, fromuser, fromdomain,
					&dkim_s_m);
			if (sm_is_err(ret)) goto error;
			if (dkim_s_m != NULL)
				dkim_s = dkim_s_m;
			else
				dkim_s = dkim_ctx->dkc_dk_s_dflt;
			SM_STR_FREE(fromuser);
			SM_STR_FREE(fromdomain);
		}
		else {
			ret = sm_error_temp(SM_EM_DKIM, EINVAL);
			goto error;
		}
	}

	/* sign it... separate function? */
	if (dkim_s != NULL && dkim_s->dks_rsa != NULL) {
		uint lc;
		uchar b64sig[256];	/* length??? */
		dkh_P dkh;

		sm_str_clr(dkim_msg->dkm_hdrs);
		DKH_FOREACH_REV(dkh, &(dkim_msg->dkm_hhd)) {
			ret = sm_str_cat(dkim_msg->dkm_hdrs, dkh->dkh_hdr_n);
			if (sm_is_err(ret)) goto error;
		}

		if (dkim_cnf->dkimcnf_add[0] != NULL) {
			ret = dkim_add_hdrs(dkim_cnf, dkim_msg);
			if (sm_is_err(ret)) goto error;
		}

		/* remove the last colon from the header list */
		ret = sm_str_rm_trail(dkim_msg->dkm_hdrs, ":");
		if (sm_is_err(ret)) goto error;

		/* maybe add some formatting here to break lines that are too long? */
		ret = sm_strprintf(dkim_msg->dkm_sighdr,
			"DKIM-Signature: v=1; a=rsa-sha%s; c=%s/%s; d=%s;\r\n\ts=%s; t=%d;\r\n\tbh=%s; h=%S;\r\n\tb=",
			dkim_cookie->dkf_sha256 ? "256" : "1",
			DKIM_CANON(dkim_cnf->dkimcnf_hcanon),
			DKIM_CANON(dkim_cnf->dkimcnf_bcanon),
			dkim_s->dks_domain,
			dkim_s->dks_selector,
			dkim_msg->dkm_time,
			dkim_msg->dkm_b64bhash, dkim_msg->dkm_hdrs);
		if (sm_is_err(ret)) goto error;

		if (DKIM_C_REL(dkim_cnf->dkimcnf_hcanon)) {
			ret = sm_io_fprintf(dkim_msg->dkm_hfp,
				"dkim-signature:v=1; a=rsa-sha%s; c=%s/%s; d=%s; s=%s; t=%d; bh=%s; h=%S; b=",
				dkim_cookie->dkf_sha256 ? "256" : "1",
				DKIM_CANON(dkim_cnf->dkimcnf_hcanon),
				DKIM_CANON(dkim_cnf->dkimcnf_bcanon),
				dkim_s->dks_domain,
				dkim_s->dks_selector,
				dkim_msg->dkm_time,
				dkim_msg->dkm_b64bhash, dkim_msg->dkm_hdrs);
		}
		else
			ret = sm_io_fputstr(dkim_msg->dkm_hfp, dkim_msg->dkm_sighdr);
		if (sm_is_err(ret)) goto error;

		dkim_gethash(dkim_msg->dkm_hfp, &hash, &hashlen);

		dkim_msg->dkm_s = dkim_s;
		ret = dkim_gensig(dkim_ctx, dkim_msg, dkim_cookie);
		if (sm_is_err(ret)) goto error;

		sm_memzero(b64sig, sizeof(b64sig));
		sm_base64_encode(dkim_msg->dkm_rsaout, dkim_msg->dkm_rsaoutlen,
			b64sig, sizeof(b64sig));

		/* note: this breaks the simple strstr() tests that look for b= */
		hashlen = strlen((const char *)b64sig);
		for (i = 0, lc = 0; i < hashlen; i++) {
			if (lc++ >= 60) {
				ret = sm_str_scat(dkim_msg->dkm_sighdr, "\r\n\t ");
				if (sm_is_err(ret)) goto error;
				lc = 0;
			}
			ret = SM_STR_PUT(dkim_msg->dkm_sighdr, b64sig[i]);
			if (sm_is_err(ret)) goto error;
		}

		ret = sm_str_scat(dkim_msg->dkm_sighdr, "\r\n");
		if (sm_is_err(ret)) goto error;

	}
	ret = dkim_close(dkim_msg->dkm_hfp, 0);
	if (sm_is_err(ret)) goto error;
	dkim_msg->dkm_hfp = NULL;
	return SM_SUCCESS;

  enomem:
	ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
  error:
	SM_STR_FREE(fromuser);
	SM_STR_FREE(fromdomain);

	/* cleanup? done in caller by freeing msg context */
	return ret;
}
#endif /* MTA_USE_DKIM */
