/*
 * Copyright (c) 2000-2005 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 * Copyright (c) 1990
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Chris Torek.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: vfprintf.c,v 1.85 2014/01/30 16:37:29 ca Exp $")
#include "sm/varargs.h"
#include "sm/assert.h"
#include "sm/io.h"
#include "sm/memops.h"
#include "sm/ctype.h"
#include "sm/string.h"
#include "sm/base64.h"
#include "sm/str.h"
#include "sm/str-int.h"
#if !SM_NO_CSTR
#include "sm/cstr.h"
#endif
#include "sm/error.h"
#include "sm/heap.h"
#include "io-int.h"
#include "print.h"
#include "fvwrite.h"
#if MTA_USE_FP_PRINT
# include <stdio.h>	/* for sprintf() */
#endif
#include "sm/net.h"

/*
**  Format specifiers
**
**  +   A       IPv4 address
**  +   a       sockaddr
**  +   B       IPv6 address
**      b       - [binary?]
**  +   C       sm_cstr
**      c       char
**      D       long int (decimal)
**      d       int (decimal)
**      E       double float (NOT IMPL)
**      e       double float (NOT IMPL)
**      F       -
**      f       float (NOT IMPL)
**      G       double float (NOT IMPL)
**      g       double float (NOT IMPL)
**  +   H       module name
**      h       [modifier]
**  +   I       ipaddr
**      i       int (decimal)
**      J       -
**      j       -
**      K       -
**      k       -
**      L       -
**      l       [modifier]
**      M       error description (do not use width/prec/...)
**      m       error text
**  +   N       sm_str (does not print trailing '\0')
**      n       number of characters written so far stored in int *
**      O       unsigned octal
**      o       octal
**      P       -
**      p       pointer
**      Q       -
**      q       LLONGINT
**      R       -
**  +   r       if <0: print in hex, otherwise decimal
**              this is very useful for sm_ret_T
**  +   S       sm_str
**      s       char *
**  +   T       sm_str without trailing \r, \n, or \0
**      t       pointer
**      U       unsigned long int
**      u       unsigned int
**      V       -
**      v       time_t
**      W       -
**      w       -
**      X       hex (upper case)
**      x       hex (lower case)
**  +   y       hex array XX:XX:XX (length: .L aka prec)
**  +   Y       base 64
**      y       -
**  +   Z       base 32
**      z       (size_t?)
**
**  Note: specifiers marked as '+' are local additions which cause gcc
**      to create bogus error messages because it can't be told about them...
**
**  C99 modifiers:
**      t       ptrdiff_t
**      z       size_t
**  Local additions:
**      #       for string types: "sanitize" string (replace unprintable
**              characters by %HEX).
**      @       for string types: same as above (#) and replace ',' with %2C
**              (useful for logging: ',' is then a unique delimiter).
**              If this is changed: grep '%@' in all files.
**      &       for string types: xtextify and replace ',' with +2C
*/

/*
**  Overall:
**  Actual printing innards.
**  This code is large and complicated...
*/

static void	sm_find_arguments(const char *, va_list, va_list **);
static void	sm_grow_type_table(int, uchar **, int *);

/*
**  SM_PRINT -- print/flush to the file
**
**  Flush out all the vectors defined by the given uio,
**  then reset it so that it can be reused.
**
**	Parameters:
**		fp -- file pointer
**		uio -- vector list of memory locations of data for printing
**
**	Results:
**		Success: 0 (zero)
**		Failure:
*/

int
sm_print(sm_file_T *fp, struct sm_uio *uio)
{
	int err;

	if (uio->uio_resid == 0) {
		uio->uio_iovcnt = 0;
		return 0;
	}
	err = sm_fvwrite(fp, uio);
	uio->uio_resid = 0;
	uio->uio_iovcnt = 0;
	return err;
}

/*
**  SM_BPRINTF -- allow formating to an unbuffered file.
**
**  Helper function for `fprintf to unbuffered unix file': creates a
**  temporary buffer (via a "fake" file pointer).
**  We only work on write-only files; this avoids
**  worries about ungetc buffers and so forth.
**
**	Parameters:
**		fp -- the file to send the o/p to
**		fmt -- format instructions for the o/p
**		ap -- vectors of data units used for formating
**
**	Results:
**		Failure: SM_IO_EOF and errno set
**		Success: number of data units used in the formating
**
**	Side effects:
**		formatted o/p can be SM_IO_BUFSIZ length maximum
*/

static int
sm_bprintf(sm_file_T *fp, const char *fmt, va_list ap)
{
	int ret;
	sm_file_T fake;
	uchar buf[SM_IO_BUFSIZ];

	/* copy the important variables */
	fake.sm_magic = SM_FILE_MAGIC;
	fake.f_timeout = SM_TIME_FOREVER;
	f_flags(fake) = f_flags(*fp) & ~SMNBF;
	f_fd(fake) = f_fd(*fp);
	fake.f_cookie = fp->f_cookie;
	f_read(fake) = NULL;
	f_write(fake) = f_write(*fp);
	f_close(fake) = NULL;
	f_open(fake) = NULL;
	f_seek(fake) = NULL;
	f_setinfo(fake) = f_getinfo(fake) = NULL;

	/* set up the buffer */
	f_bfbase(fake) = f_p(fake) = buf;
	f_bfsize(fake) = f_w(fake) = sizeof(buf);

	/* do the work, then copy any error status */
	ret = sm_io_vfprintf(&fake, fmt, ap);
	if (ret >= 0 && sm_io_flush(&fake))
		ret = SM_IO_EOF;	/* errno set by sm_io_flush */
	if (f_flags(fake) & SMERR) {
		f_flags(*fp) |= SMERR;
#if SM_IO_ERR_VAL
		f_error(*fp) = f_error(fake);
#endif
	}
	return ret;
}


#define STATIC_ARG_TBL_SIZE 8	/* Size of static argument table. */

/*
**  SM_IO_VFPRINTF -- performs actual formating for o/p
**
**	Parameters:
**		fp -- file pointer for o/p
**		timeout -- time to complete the print
**		fmt0 -- formating directives
**		ap -- vectors with data units for formating
**
**	Results:
**		Success: number of data units used for formatting
**		Failure: SM_IO_EOF and sets errno
*/

int
sm_io_vfprintf(sm_file_T *fp, const char *fmt0, va_list ap)
{
	const char *fmt;       /* format string */
	int ch;                /* character from fmt */
	int n, m, n2;          /* handy integers (short term usage) */
	const char *cp;        /* handy const char pointer (short term usage) */
	char *wcp;             /* handy char pointer (short term usage) */
	struct sm_iov *iovp;   /* for VFP_PRINT macro */
	int flags;             /* flags as above */
	int ret;               /* return value accumulator */
	int width;             /* width from format (%8d), or 0 */
	int prec;              /* precision from format (%.3d), or -1 */
	char *encode; /* characters to be encoded, first must be encoding prefix */
	char sign;             /* sign prefix (' ', '+', '-', or \0) */
	wchar_t wc;
	ulonglong_T _uquad;    /* integer arguments %[diouxX] */
	enum { OCT, DEC, HEX, BASE32, BASE64 } base;/* base for [diouxX] conversion */
	int dprec;             /* a copy of prec if [diouxX], 0 otherwise */
	int realsz;            /* field size expanded by dprec */
	int size;              /* size of converted field or string */
	int osize;             /* original size before sanitizing */
	char *xdigs="0123456789abcdef"; /* digits for [xX] conversion */
#define SM_B32DIGS "0123456789ABCDEFGHIJKLMNOPQRSTUV"
	const char *b64digs = SM_B64DIGS;
	const char *b32digs = SM_B32DIGS;
#define NIOV 8
	struct sm_uio uio;     /* output information: summary */
	struct sm_iov iov[NIOV];/* ... and individual io vectors */
	char buf[BUF];         /* space for %c, %[diouxX], %[eEfgG] */
	char ox[2];            /* space for 0x hex-prefix */
	va_list *argtable;     /* args, built due to positional arg */
	va_list statargtable[STATIC_ARG_TBL_SIZE];
	int nextarg;           /* 1-based argument index */
	sm_str_P str;
#if !SM_NO_CSTR
	sm_cstr_P cstr;
#endif
	va_list orgap;         /* original argument pointer */

	/*
	**  Choose PADSIZE to trade efficiency vs. size.  If larger printf
	**  fields occur frequently, increase PADSIZE and make the initialisers
	**  below longer.
	*/
#define PADSIZE	16		/* pad chunk size */
	static char blanks[PADSIZE] =
	 {' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '};
	static char zeroes[PADSIZE] =
	 {'0','0','0','0','0','0','0','0','0','0','0','0','0','0','0','0'};

	/*
	**  Get * arguments, including the form *nn$.  Preserve the nextarg
	**  that the argument can be gotten once the type is determined.
	*/
#define GETASTER(val) \
	n2 = 0; \
	cp = fmt; \
	while (is_digit(*cp)) { \
		n2 = 10 * n2 + to_digit(*cp); \
		cp++; \
	} \
	if (*cp == '$') { \
		int hold = nextarg; \
		if (argtable == NULL) { \
			argtable = statargtable; \
			sm_find_arguments(fmt0, orgap, &argtable); \
		} \
		nextarg = n2; \
		val = GETARG(int); \
		nextarg = hold; \
		fmt = ++cp; \
	} \
	else { \
		val = GETARG(int); \
	}

/*
**  Get the argument indexed by nextarg.   If the argument table is
**  built, use it to get the argument.  If its not, get the next
**  argument (and arguments must be gotten sequentially).
*/

# define GETARG(type) \
	(((argtable != NULL) ? (void) (ap = argtable[nextarg]) : (void) 0), \
	 nextarg++, va_arg(ap, type))

	/* sorry, fprintf(read_only_file, "") returns SM_IO_EOF, not 0 */
	if (cantwrite(fp)) {
		errno = EBADF;
		return SM_IO_EOF;
	}

	/* optimise fprintf(stderr) (and other unbuffered Unix files) */
	if ((f_flags(*fp) & (SMNBF|SMWR|SMRW)) == (SMNBF|SMWR) &&
	    f_fd(*fp) >= 0)
		return sm_bprintf(fp, fmt0, ap);

	fmt = fmt0;
	argtable = NULL;
	nextarg = 1;
	SM_VA_COPY(orgap, ap);
	uio.uio_iov = iovp = iov;
	uio.uio_resid = 0;
	uio.uio_iovcnt = 0;
	ret = 0;

	/* Scan the format for conversions (`%' character). */
	for (;;) {
		cp = fmt;
		n = 0;
		encode = "";
		while ((wc = *fmt) != '\0') {
			if (wc == '%') {
				n = 1;
				break;
			}
			fmt++;
		}
		if ((m = fmt - cp) != 0) {
			VFP_PRINT(cp, m);
			ret += m;
		}
		if (n <= 0)
			goto done;
		fmt++;		/* skip over '%' */

		flags = 0;
		dprec = 0;
		width = 0;
		prec = -1;
		sign = '\0';
		osize = 0;	/* bogus "might be used uninitialized" */

rflag:		ch = *fmt++;
reswitch:	switch (ch)
		{
		  case ' ':

			/*
			**  ``If the space and + flags both appear, the space
			**  flag will be ignored.''
			**	-- ANSI X3J11
			*/

			if (!sign)
				sign = ' ';
			goto rflag;
		  case '#':
			flags |= ALT;
			encode = "%";
			goto rflag;
		  case '&':
			flags |= ALT;
			encode = "+, <>\"\\()";
/*
(c < '!' || c > '~' || c == '+' || c == '\\' || c == '(' || strchr(taboo, c)
xtextify((s), "<>\")", str);	\
*/
			goto rflag;
		  case '@':
			flags |= ALT;
			encode = "%,";
			goto rflag;
		  case '*':

			/*
			**  ``A negative field width argument is taken as a
			**  - flag followed by a positive field width.''
			**	-- ANSI X3J11
			**  They don't exclude field widths read from args.
			*/

			GETASTER(width);
			if (width >= 0)
				goto rflag;
			width = -width;
			/* FALLTHROUGH */
		  case '-':
			flags |= LADJUST;
			goto rflag;
		  case '+':
			sign = '+';
			goto rflag;
		  case '.':
			if ((ch = *fmt++) == '*') {
				GETASTER(n);
				prec = n < 0 ? -1 : n;
				goto rflag;
			}
			n = 0;
			while (is_digit(ch)) {
				n = 10 * n + to_digit(ch);
				ch = *fmt++;
			}
			if (ch == '$') {
				nextarg = n;
				if (argtable == NULL) {
					argtable = statargtable;
					sm_find_arguments(fmt0, orgap, &argtable);
				}
				goto rflag;
			}
			prec = n < 0 ? -1 : n;
			goto reswitch;
		  case '0':

			/*
			**  ``Note that 0 is taken as a flag, not as the
			**  beginning of a field width.''
			**	-- ANSI X3J11
			*/

			flags |= ZEROPAD;
			goto rflag;
		  case '1': case '2': case '3': case '4':
		  case '5': case '6': case '7': case '8': case '9':
			n = 0;
			do {
				n = 10 * n + to_digit(ch);
				ch = *fmt++;
			} while (is_digit(ch));
			if (ch == '$') {
				nextarg = n;
				if (argtable == NULL) {
					argtable = statargtable;
					sm_find_arguments(fmt0, orgap, &argtable);
				}
				goto rflag;
			}
			width = n;
			goto reswitch;
		  case 'h':
			if (*fmt == 'h') {
				fmt++;
				flags |= CHARINT;
			} else {
				flags |= SHORTINT;
			}
			goto rflag;
		  case 'l':
			if (*fmt == 'l') {
				fmt++;
				flags |= LLONGINT;
			}
			else
				flags |= LONGINT;
			goto rflag;
		  case 'q':
			flags |= LLONGINT;
			goto rflag;
		  case 't':
			flags |= PTRINT;
			goto rflag;
		  case 'z':
			flags |= SIZEINT;
			goto rflag;
		  case 'e':
		  case 'E':
		  case 'f':
		  case 'g':
		  case 'G':
#if MTA_USE_FP_PRINT
			{
				double val;
				char *p;
				char fmt[16];
				char out[150];
				size_t len;

				/*
				**  This code implements floating point output
				**  in the most portable manner possible,
				**  relying only on 'sprintf' as defined by
				**  the 1989 ANSI C standard.
				**  We silently cap width and precision
				**  at 120, to avoid buffer overflow.
				*/

				val = GETARG(double);

				p = fmt;
				*p++ = '%';
				if (sign)
					*p++ = sign;
				if (flags & ALT)
					*p++ = '#';
				if (flags & LADJUST)
					*p++ = '-';
				if (flags & ZEROPAD)
					*p++ = '0';
				*p++ = '*';
				if (prec >= 0) {
					*p++ = '.';
					*p++ = '*';
				}
				*p++ = ch;
				*p = '\0';

				if (width > 120)
					width = 120;
				if (prec > 120)
					prec = 120;
				if (prec >= 0)
					sprintf(out, fmt, width, prec, val);
				else
					sprintf(out, fmt, width, val);
				len = strlen(out);
				VFP_PRINT(out, len);
				VFP_FLUSH();
				continue;
			}
#else /* MTA_USE_FP_PRINT */

		/* fixme: Complain about unsupported format?? */
#define FPNOTSUPP "FPnotsupp"
			VFP_PRINT(FPNOTSUPP, strlen(FPNOTSUPP));
			VFP_FLUSH();
			continue;

#endif /* MTA_USE_FP_PRINT */
		  case 'n':
			if (flags & LLONGINT)
				*GETARG(longlong_T *) = ret;
			else if (flags & LONGINT)
				*GETARG(long *) = ret;
			else if (flags & SHORTINT)
				*GETARG(short *) = ret;
			else if (flags & CHARINT)
				*GETARG(signed char *) = ret;
			else if (flags & PTRINT)
				*GETARG(ptrdiff_t *) = ret;
			else if (flags & SIZEINT)
				*GETARG(ssize_t *) = ret;
			else
				*GETARG(int *) = ret;
			continue;	/* no output */

#include "print.c"

		  default:	/* "%?" prints ?, unless ? is NUL */
			if (ch == '\0')
				goto done;
			/* pretend it was %c with argument ch */
			*buf = ch;
			cp = buf;
			size = 1;
			sign = '\0';
			break;
		}

#include "printit.c"
	}
done:
	VFP_FLUSH();
error:
	va_end(orgap);
	if ((argtable != NULL) && (argtable != statargtable))
		sm_free(argtable);
	return sm_error(fp) ? SM_IO_EOF : ret;
	/* NOTREACHED */
}

/*
 * Type ids for argument type table.
 */
#define T_UNUSED	0
#define T_SHORT		1
#define T_U_SHORT	2
#define TP_SHORT	3
#define T_INT		4
#define T_U_INT		5
#define TP_INT		6
#define T_LONG		7
#define T_U_LONG	8
#define TP_LONG		9
#define T_LLONG		10
#define T_U_LLONG	11
#define TP_LLONG	12
#define T_DOUBLE	13
#define T_LONG_DOUBLE	14
#define TP_CHAR		15
#define TP_VOID		16
#define T_PTRINT	17
#define TP_PTRINT	18
#define T_SIZEINT	19
#define T_SSIZEINT	20
#define TP_SSIZEINT	21
#define T_MAXINT	22
#define T_MAXUINT	23
#define TP_MAXINT	24
#define T_CHAR		25
#define T_U_CHAR	26
#define TP_SM_STR	27
#define TP_SM_CSTR	28
#define TP_SM_IPV6	29

/*
**  SM_FIND_ARGUMENTS -- find all args when a positional parameter is found.
**
**  Find all arguments when a positional parameter is encountered.  Returns a
**  table, indexed by argument number, of pointers to each arguments.  The
**  initial argument table should be an array of STATIC_ARG_TBL_SIZE entries.
**  It will be replaced with a malloc-ed one if it overflows.
**
**	Parameters:
**		fmt0 -- formating directives
**		ap -- vector list of data unit for formating consumption
**		argtable -- an indexable table (returned) of 'ap'
**
**	Results:
**		none.
*/

static void
sm_find_arguments(const char *fmt0, va_list ap, va_list **pargtable)
{
	const char *fmt; /* format string */
	int ch;		/* character from fmt */
	int n, n2;	/* handy integer (short term usage) */
	const char *cp;	/* handy char pointer (short term usage) */
	int flags;	/* flags as above */
	uchar *typetable; /* table of types */
	uchar stattypetable[STATIC_ARG_TBL_SIZE];
	int tablesize;		/* current size of type table */
	int tablemax;		/* largest used index in table */
	int nextarg;		/* 1-based argument index */

	/* Add an argument type to the table, expanding if necessary. */
#define ADDTYPE(type) \
	((nextarg >= tablesize) ? \
		(sm_grow_type_table(nextarg, &typetable, &tablesize), 0) : 0, \
	(nextarg > tablemax) ? tablemax = nextarg : 0, \
	typetable[nextarg++] = type)

#define	ADDSARG() \
        ((flags&MAXINT) ? ADDTYPE(T_MAXINT) : \
	    ((flags&PTRINT) ? ADDTYPE(T_PTRINT) : \
	    ((flags&SIZEINT) ? ADDTYPE(T_SSIZEINT) : \
	    ((flags&LLONGINT) ? ADDTYPE(T_LLONG) : \
	    ((flags&LONGINT) ? ADDTYPE(T_LONG) : \
	    ((flags&SHORTINT) ? ADDTYPE(T_SHORT) : \
	    ((flags&CHARINT) ? ADDTYPE(T_CHAR) : ADDTYPE(T_INT))))))))

#define	ADDUARG() \
        ((flags&MAXINT) ? ADDTYPE(T_MAXUINT) : \
	    ((flags&PTRINT) ? ADDTYPE(T_PTRINT) : \
	    ((flags&SIZEINT) ? ADDTYPE(T_SIZEINT) : \
	    ((flags&LLONGINT) ? ADDTYPE(T_U_LLONG) : \
	    ((flags&LONGINT) ? ADDTYPE(T_U_LONG) : \
	    ((flags&SHORTINT) ? ADDTYPE(T_U_SHORT) : \
	    ((flags&CHARINT) ? ADDTYPE(T_U_CHAR) : ADDTYPE(T_U_INT))))))))


	/* Add * arguments to the type array. */
#define ADDASTER() \
	n2 = 0; \
	cp = fmt; \
	while (is_digit(*cp)) { \
		n2 = 10 * n2 + to_digit(*cp); \
		cp++; \
	} \
	if (*cp == '$') { \
		int hold = nextarg; \
		nextarg = n2; \
		ADDTYPE (T_INT); \
		nextarg = hold; \
		fmt = ++cp; \
	} \
	else { \
		ADDTYPE (T_INT); \
	}

	fmt = fmt0;
	typetable = stattypetable;
	tablesize = STATIC_ARG_TBL_SIZE;
	tablemax = 0;
	nextarg = 1;
	(void) sm_memset(typetable, T_UNUSED, STATIC_ARG_TBL_SIZE);

	/* Scan the format for conversions (`%' character). */
	for (;;) {
		for (cp = fmt; (ch = *fmt) != '\0' && ch != '%'; fmt++)
			/* void */;
		if (ch == '\0')
			goto done;
		fmt++;		/* skip over '%' */

		flags = 0;

rflag:		ch = *fmt++;
reswitch:	switch (ch)
		{
		  case ' ':
		  case '#':
			goto rflag;
		  case '*':
			ADDASTER();
			goto rflag;
		  case '-':
		  case '+':
			goto rflag;
		  case '.':
			if ((ch = *fmt++) == '*') {
				ADDASTER();
				goto rflag;
			}
			while (is_digit(ch))
				ch = *fmt++;
			goto reswitch;
		  case '0':
			goto rflag;
		  case '1': case '2': case '3': case '4':
		  case '5': case '6': case '7': case '8': case '9':
			n = 0;
			do {
				n = 10 * n + to_digit(ch);
				ch = *fmt++;
			} while (is_digit(ch));
			if (ch == '$') {
				nextarg = n;
				goto rflag;
			}
			goto reswitch;
		  case 'h':
			flags |= SHORTINT;
			goto rflag;
		  case 'l':
			if (*fmt == 'l') {
				fmt++;
				flags |= LLONGINT;
			}
			else
				flags |= LONGINT;
			goto rflag;
		  case 'q':
			flags |= LLONGINT;
			goto rflag;
		  case 't':
			flags |= PTRINT;
			goto rflag;
		  case 'z':
			flags |= SIZEINT;
			goto rflag;
		  case 'c':
			ADDTYPE(T_INT);
			break;
		  case 'D':
			flags |= LONGINT;
			/*FALLTHROUGH*/
		  case 'd':
		  case 'i':
		  case 'r':
		  case 'w':
			if (flags & LLONGINT)
				ADDTYPE(T_LLONG);
			else if (flags & PTRINT)
				ADDTYPE(T_PTRINT);
			else if (flags & SIZEINT)
				ADDTYPE(T_SSIZEINT);
			else
				ADDSARG();
			break;
		  case 'e':
		  case 'E':
		  case 'f':
		  case 'g':
		  case 'G':
			ADDTYPE(T_DOUBLE);
			break;
		  case 'n':
			if (flags & LLONGINT)
				ADDTYPE(TP_LLONG);
			else if (flags & LONGINT)
				ADDTYPE(TP_LONG);
			else if (flags & SHORTINT)
				ADDTYPE(TP_SHORT);
			else if (flags & PTRINT)
				ADDTYPE(TP_PTRINT);
			else if (flags & SIZEINT)
				ADDTYPE(TP_SSIZEINT);
			else
				ADDTYPE(TP_INT);
			continue;	/* no output */
		  case 'O':
			flags |= LONGINT;
			/*FALLTHROUGH*/
		  case 'o':
			if (flags & LLONGINT)
				ADDTYPE(T_U_LLONG);
			else
				ADDUARG();
			break;
		  case 'p':
			ADDTYPE(TP_VOID);
			break;
		  case 's':
			ADDTYPE(TP_CHAR);
			break;
		  case 'U':
			flags |= LONGINT;
			/*FALLTHROUGH*/
		  case 'u':
			if (flags & LLONGINT)
				ADDTYPE(T_U_LLONG);
			else
				ADDUARG();
			break;
		  case 'X':
		  case 'x':
		  case 'Y':
		  case 'Z':
			if (flags & LLONGINT)
				ADDTYPE(T_U_LLONG);
			else if (flags & PTRINT)
				ADDTYPE(T_PTRINT);
			else if (flags & SIZEINT)
				ADDTYPE(T_SIZEINT);
			else
				ADDUARG();
			break;
		  case 'A':
			/* deal with 'l' modifier? */
#if SIZEOF_INT == 4
			ADDTYPE(T_U_INT);
#elif SIZEOF_LONG == 4
			ADDTYPE(T_U_LONG);
#else
#endif
			break;
		  case 'a':
#if MTA_NETINET6
			ADDTYPE(TP_SM_IPV6);
#endif
			break;
#if !SM_NO_CSTR
		  case 'C':
			ADDTYPE(TP_SM_CSTR);
			break;
#endif /* !SM_NO_CSTR */
		  case 'N':
		  case 'S':
		  case 'T':
			ADDTYPE(TP_SM_STR);
			break;
		  default:	/* "%?" prints ?, unless ? is NUL */
			if (ch == '\0')
				goto done;
			break;
		}
	}
done:
	/* Build the argument table. */
	if (tablemax >= STATIC_ARG_TBL_SIZE) {
		SM_ASSERT(pargtable != NULL);
		*pargtable = (va_list *)
		    sm_malloc(sizeof(va_list) * (tablemax + 1));
/* XXX check result? */
		SM_ASSERT(*pargtable != NULL);
	}

#if 0
	(*pargtable)[0] = NULL;
#endif /* 0 */
	for (n = 1; n <= tablemax; n++) {
		SM_VA_COPY((*pargtable)[n], ap);
		switch (typetable[n]) {
		  case T_UNUSED:
			(void) va_arg(ap, int);
			break;
		  case T_SHORT:
			(void) va_arg(ap, int);
			break;
		  case T_U_SHORT:
			(void) va_arg(ap, int);
			break;
		  case TP_SHORT:
			(void) va_arg(ap, short *);
			break;
		  case T_INT:
			(void) va_arg(ap, int);
			break;
		  case T_U_INT:
			(void) va_arg(ap, uint);
			break;
		  case TP_INT:
			(void) va_arg(ap, int *);
			break;
		  case T_LONG:
			(void) va_arg(ap, long);
			break;
		  case T_U_LONG:
			(void) va_arg(ap, ulong);
			break;
		  case TP_LONG:
			(void) va_arg(ap, long *);
			break;
		  case T_LLONG:
			(void) va_arg(ap, longlong_T);
			break;
		  case T_U_LLONG:
			(void) va_arg(ap, ulonglong_T);
			break;
		  case TP_LLONG:
			(void) va_arg(ap, longlong_T *);
			break;
		  case T_DOUBLE:
			(void) va_arg(ap, double);
			break;
		  case TP_CHAR:
			(void) va_arg(ap, char *);
			break;
		  case TP_VOID:
			(void) va_arg(ap, void *);
			break;
		  case T_PTRINT:
			(void) va_arg(ap, ptrdiff_t);
			break;
		  case TP_PTRINT:
			(void) va_arg(ap, ptrdiff_t *);
			break;
		  case T_SIZEINT:
			(void) va_arg(ap, size_t);
			break;
		  case T_SSIZEINT:
			(void) va_arg(ap, ssize_t);
			break;
		  case TP_SSIZEINT:
			(void) va_arg(ap, ssize_t *);
			break;
		}
	}

	if ((typetable != NULL) && (typetable != stattypetable))
		sm_free(typetable);
}

/*
**  SM_GROW_TYPE_TABLE -- Increase the size of the type table.
**
**	Parameters:
**		tabletype -- type of table to grow
**		tablesize -- requested new table size
*/

static void
sm_grow_type_table(int nextarg, uchar **typetable, int *tablesize)
{
	uchar *oldtable = *typetable;
	int newsize = *tablesize * 2;

	if (newsize < nextarg + 1)
		newsize = nextarg + 1;
	if (*tablesize == STATIC_ARG_TBL_SIZE) {
		*typetable = (uchar *) sm_malloc(sizeof(uchar)
							   * newsize);
/* XXX check result? */
		SM_ASSERT(*typetable != NULL);
		(void) sm_memmove(*typetable, oldtable, *tablesize);
	}
	else {
		*typetable = (uchar *) sm_realloc(*typetable,
					sizeof(uchar) * newsize);
/* XXX check result? */
		SM_ASSERT(*typetable != NULL);
	}
	(void) sm_memset(*typetable + *tablesize, T_UNUSED,
		       (newsize - *tablesize));

	*tablesize = newsize;
}
