/*
 * Copyright (c) 2014 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: testmap.c,v 1.5 2014/06/01 00:50:21 ca Exp $")

#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/heap.h"
#include "sm/fcntl.h"
#include "map.h"
#include "sm/map.h"
#include "sm/maps.h"
#include "sm/mapc.h"
#include "sm/mapclasses.h"

#if MTA_TEST_MAP

/*
**  a test map that can return errors
*/

/*
// static sm_map_open_F	 sm_tst_open;
// static sm_map_close_F	 sm_tst_close;
*/

typedef struct sm_tstmap_S sm_tstmap_T, *sm_tstmap_P;
struct sm_tstmap_S
{
	int		 	 tstmap_tb4ok;
	int		 	 tstmap_tb4nomore;
	int		 	 tstmap_ctb4nomore;
	sm_str_P 	 tstmap_stb4nomore;
};

/*
**  These are types of databases.
*/

#define SMDB_TYPE_DEFAULT	NULL
#define SMDB_TYPE_TST		"tst"

/*
**  SM_TST_SETOPT -- set options for map
**
**	Parameters:
**		map -- map
**		ap -- options
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_setopt(sm_map_P map, va_list ap)
{
	sm_ret_T ret;
	int k;

	SM_IS_MAP(map);
	ret = SM_SUCCESS;

	for (;;) {
		k = va_arg(ap, int);
		if (SMPO_END == k)
			break;

	/* TODO: options to set counters tb4... */

		switch (k) {
		  default:
			/* silently ignore bogus options? */
			break;
		}
	}
	return ret;
}

/*
**  SM_TST_GETOPT -- get options for map
**
**	Parameters:
**		map -- map
**		which -- which option?
**		valp -- pointer to place where result should be stored
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_getopt(sm_map_P map, int which, void *valp)
{
	sm_ret_T ret;
	sm_tstmap_P tst;

	SM_IS_MAP(map);
	tst = (sm_tstmap_P) map->sm_map_db;
	if (NULL == tst)
		return sm_error_perm(SM_EM_MAP, ENOENT);

	/* ... */

	ret = SM_SUCCESS;
	return ret;
}

/*
**  SM_TST_CLOSE -- close map
**	more parameters???
**
**	Parameters:
**		map -- map
**		flags -- currently ignored
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_close(sm_map_P map, uint32_t flags)
{
	sm_ret_T ret;
	int fd;
	sm_mapc_P mapc;
	sm_tstmap_P sm_tstmap;

	SM_IS_MAP(map);
	mapc = map->sm_map_class;
	SM_IS_MAPC(mapc);
	ret = SM_SUCCESS;

	sm_tstmap = (sm_tstmap_P) map->sm_map_db;
	if (NULL == sm_tstmap)
		return sm_error_perm(SM_EM_MAP, ENOENT);

	sm_free_size(sm_tstmap, sizeof(sm_tstmap));

	/* close TST map */
	map->sm_map_db = NULL;

	return ret;
}

/*
**  SM_TST_DESTROY -- destroy map
**	XXX more parameters...
**
**	Parameters:
**		map -- map
**		flags -- flags for map
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_destroy(sm_map_P map, uint32_t flags)
{
	sm_ret_T ret;
	sm_mapc_P mapc;
	sm_tstmap_P sm_tstmap;

	SM_IS_MAP(map);
	mapc = map->sm_map_class;
	SM_IS_MAPC(mapc);
	ret = SM_SUCCESS;

	sm_tstmap = (sm_tstmap_P) map->sm_map_db;
	if (NULL == sm_tstmap)
		return sm_error_perm(SM_EM_MAP, ENOENT);

	/* close fd? */
	sm_free_size(sm_tstmap, sizeof(sm_tstmap));

	map->sm_map_db = NULL;
	return ret;
}

/*
**  SM_TST_CREATE -- create map
**
**	Parameters:
**		mapc -- map context
**		type -- type of map
**		flags -- flags for map
**		map -- map
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_create(sm_mapc_P mapc, const sm_cstr_P type, uint32_t flags, sm_map_P map)
{
	sm_tstmap_P sm_tstmap;

	SM_IS_MAPC(mapc);
	SM_REQUIRE(map != NULL);
	sm_tstmap = (sm_tstmap_P) sm_zalloc(sizeof(*sm_tstmap));
	if (NULL == sm_tstmap)
		return sm_error_temp(SM_EM_MAP, ENOMEM);
	sm_tstmap->tstmap_tb4ok = 1;
	sm_tstmap->tstmap_tb4nomore = 1;
	sm_tstmap->tstmap_ctb4nomore = 1;
	sm_tstmap->tstmap_stb4nomore = sm_str_scpy0(NULL, "smtpc_rcpt_conf:mn", 64);
	if (NULL == sm_tstmap->tstmap_stb4nomore)
		return sm_error_temp(SM_EM_MAP, ENOMEM);
	map->sm_map_db = sm_tstmap;
	map->sm_map_caps = SMMAP_CAPS_LTALL;
	return SM_SUCCESS;
}

/*
**  SM_TST_OPEN -- open map
**
**	Parameters:
**		mapc -- map context
**		type -- type of map
**		flags -- flags for opening map
**		path -- path of map
**		mode -- open mode (currently ignored, DBMMODE is used)
**		map -- map
**		ap -- additional argument
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_open(sm_mapc_P mapc, const sm_cstr_P type, uint32_t flags, const char *path, int mode, sm_map_P map, va_list ap)
{
	SM_IS_MAPC(mapc);
	return SM_SUCCESS;
}

/*
**  SM_TST_LOOKUP -- lookup a key in TST, return data if found
**
**	Parameters:
**		map -- map context
**		flags -- flags
**		key -- key
**		data -- data (output)
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
sm_tst_lookup(sm_map_P map, uint32_t flags, sm_map_key_P key, sm_map_data_P data)
{
	sm_ret_T ret;
	uint idx;
	sm_mapc_P mapc;
	sm_tstmap_P sm_tstmap;

	SM_IS_MAP(map);
	SM_IS_KEY(key);
	SM_IS_DATA(data);
	mapc = map->sm_map_class;
	SM_IS_MAPC(mapc);
	ret = SM_SUCCESS;

	sm_tstmap = (sm_tstmap_P) map->sm_map_db;
	if (NULL == sm_tstmap )
		return sm_error_perm(SM_EM_MAP, ENOENT);	/* XXX */

	if (sm_tstmap->tstmap_stb4nomore != NULL &&
	    sm_str_cmpn(key, sm_tstmap->tstmap_stb4nomore, sm_str_getlen(sm_tstmap->tstmap_stb4nomore)) == 0)
	{
		if (sm_tstmap->tstmap_ctb4nomore > 0) {
			--sm_tstmap->tstmap_ctb4nomore;
			return SM_MAP_TEMPMAP;
		}
		return sm_error_perm(SM_EM_MAP, SM_E_NOMORE);
	}

	ret = sm_str_f1chr(key, ':', 0, &idx);
	if (sm_is_err(ret))
		idx = 0;
	else {
		++idx;
		if (idx < sm_str_getlen(key) && sm_str_rd_elem(key, idx) == 'm')
			++idx;
	}

	if (idx >= sm_str_getlen(key))
		return SM_MAP_PERMMAP;
	if (sm_str_getlen(key) < 1)
		return SM_MAP_NOTFOUND;

	if (sm_str_rd_elem(key, idx) == '4')
		return SM_MAP_TEMPMAP;
	if (sm_str_rd_elem(key, idx) == '5')
		return SM_MAP_PERMMAP;
	if (sm_str_rd_elem(key, idx) == '3')
		return SM_MAP_DATA2BIG;
	if (sm_str_rd_elem(key, idx) == '6')
		return SM_MAP_NOTFOUND;
	if (sm_str_rd_elem(key, idx) == 'o') {
		if (sm_tstmap->tstmap_tb4ok > 0) {
			--sm_tstmap->tstmap_tb4ok;
			return SM_MAP_TEMPMAP;
		}
	}
	if (sm_str_rd_elem(key, idx) == 'n') {
		if (sm_tstmap->tstmap_tb4nomore > 0) {
			--sm_tstmap->tstmap_tb4nomore;
			return SM_MAP_TEMPMAP;
		}
		return sm_error_perm(SM_EM_MAP, SM_E_NOMORE);
	}

	ret = sm_str_catpart(data, (const sm_rdstr_P) key, idx, sm_str_getlen(key) - 1);
	if (sm_is_err(ret))
		ret = SM_MAP_PERMMAP;
	return ret;
}

/*
**  SM_TST_CLASS_CREATE -- create TST map class
**
**	Parameters:
**		maps -- map system context
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
sm_tst_class_create(sm_maps_P maps)
{
	sm_ret_T ret;
	sm_mapc_P mapc;
	sm_cstr_P htype;

#define TST_TYPE	"tst"

	ret = SM_SUCCESS;
	mapc = NULL;
	htype = sm_cstr_scpyn0((const uchar *)TST_TYPE, strlen(TST_TYPE));
	if (NULL == htype)
		goto error;

	ret = sm_mapc_create(maps, htype,
			SMMAPC_FL_GEN_REOPEN|SMMAPC_FL_LCK_FULL|
			SMMAPC_FL_FILE,
			sm_tst_create,
			sm_tst_open,
			sm_tst_close,
			NULL,
			sm_tst_destroy,
			NULL,
			NULL,
			NULL,	/* sm_tst_alloc	*/
			NULL,	/* sm_tst_free	*/
			sm_tst_lookup,
			NULL,	/* sm_tst_locate */
			NULL,	/* sm_tst_first */
			NULL,	/* sm_tst_next */
			sm_tst_setopt,
			sm_tst_getopt,
			&mapc);

	SM_CSTR_FREE(htype);
	return ret;

  error:
	if (SM_SUCCESS == ret)
		ret = sm_error_temp(SM_EM_MAP, ENOMEM);
	/* cleanup mapc? */
	return ret;
}
#endif /* MTA_TEST_MAP */
